/*
 * Copyright (C) Mellanox Technologies Ltd. 2021.  ALL RIGHTS RESERVED.
 *
 * See file LICENSE for terms.
 */

#include <string>
#include <iostream>
#include <list>
#include <tuple>

#include "imports/prometheus_aggr/prometheus_aggr.h"
#include "api/clx_api.h"

typedef std::pair<const char*, const char*> pinfo;

#define MAX_UINT32 ((uint32_t)-1)
#define MAX_UINT8 ((uint8_t)-1)
#define DEFAULT_POLLING_INTERVAL 5  // sec

#define DEFAULT_SOURCE_ID "prometheus_aggr"
#define DEFAULT_SOURCE_TAG "prometheus_aggr"
#define DEFAULT_BUFFER_SIZE (128 * 1024)       // 128KB
#define DEFAULT_FILE_SIZE (1 * 1024 * 1024)    // 1MB
#define DEFAULT_FILE_AGE (60 * 60 * 1000000L)  // 1 hour
#define DEFAULT_DATA_ROOT "prometheus_aggr"
#define DEFAULT_SCHEMA_SUBDIR "schema"
#define DEFAULT_IPC_SOCKETS_DIR "/tmp/ipc_sockets"

#ifndef UNREFERENCED_PARAMETER
#define UNREFERENCED_PARAMETER(x) (void)(x)
#endif

static bool _dummy_provider_initialize(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    UNREFERENCED_PARAMETER(ctx);
    UNREFERENCED_PARAMETER(provider);
    return true;
}

static void _dummy_provider_finalize(clx_api_context_t* ctx, clx_api_provider_t* provider) {
    UNREFERENCED_PARAMETER(ctx);
    UNREFERENCED_PARAMETER(provider);
}

static bool set_str(char*& p, const char* val) {
    if (p) {
        free(p);
    }
    p = strdup(val);
    return p != nullptr;
}

class ClxApiEventWriter: public PrometheusAggr::EventWriter {
 public:
    ClxApiEventWriter();
    ~ClxApiEventWriter();

    void SetFileWriterMaxFileSize(size_t max_file_size);
    void SetFileWriterMaxFileAge(clx_api_timestamp_t max_file_age);
    void SetFileWriterMaxDataRoot(const char* data_root);
    void EnableFileWriter(bool enable);
    void SetIPCSocketsDir(const char* ipc_sockets_dir);
    void EnableIPC(bool enable);
    void SetIPCReattachMaxInterval(uint32_t ipc_max_reattach_time_msec);
    void SetIPCReattachMaxTries(uint8_t ipc_max_reattach_tries);
    void SetIPCSocketTimeout(uint32_t ipc_socket_timeout_msec);


    bool Start(void);
    void Stop(void);

    virtual bool Write(const clx_guid_t app_id, uint64_t user_defined1, uint64_t user_defined2, const void* data,
                       uint32_t data_size);
 protected:
   void*            api_ctx_;
   clx_api_params_t api_params_;
};

ClxApiEventWriter::ClxApiEventWriter() {
    memset(&api_params_, 0, sizeof(api_params_));

    set_str(api_params_.source_id, DEFAULT_SOURCE_ID);
    set_str(api_params_.source_tag, DEFAULT_SOURCE_TAG);
    api_params_.buffer_size   = DEFAULT_BUFFER_SIZE;
    api_params_.max_file_size = DEFAULT_FILE_SIZE;
    api_params_.max_file_age  = DEFAULT_FILE_AGE;
    set_str(api_params_.data_path_template, CLX_API_DATA_PATH_TEMPLATE);
    set_str(api_params_.data_root, DEFAULT_DATA_ROOT);
    set_str(api_params_.schema_path, DEFAULT_DATA_ROOT "/" DEFAULT_SCHEMA_SUBDIR);
    set_str(api_params_.ipc_sockets_dir, DEFAULT_IPC_SOCKETS_DIR);
    api_params_.enable_opaque_events = true;
}

ClxApiEventWriter::~ClxApiEventWriter() {
    free(api_params_.source_id);
    free(api_params_.source_tag);
    free(api_params_.data_path_template);
    free(api_params_.data_root);
    free(api_params_.schema_path);
    free(api_params_.ipc_sockets_dir);
}

bool ClxApiEventWriter::Write(const clx_guid_t app_id, uint64_t user_defined1, uint64_t user_defined2, const void* data,
                              uint32_t data_size) {
    return clx_api_opaque_event_write(api_ctx_, app_id, user_defined1, user_defined2, data, data_size);
}

void ClxApiEventWriter::SetFileWriterMaxFileSize(size_t max_file_size) {
    api_params_.max_file_size = max_file_size;
}

void ClxApiEventWriter::SetFileWriterMaxFileAge(clx_api_timestamp_t max_file_age) {
    api_params_.max_file_age = max_file_age;
}

void ClxApiEventWriter::SetFileWriterMaxDataRoot(const char* data_root) {
    if (!set_str(api_params_.data_root, data_root)) {
        throw std::runtime_error("cannot duplicate data_root");
    }

    std::string schema_path = (std::string)data_root + "/" + DEFAULT_SCHEMA_SUBDIR;
    if (!set_str(api_params_.schema_path, schema_path.c_str())) {
        throw std::runtime_error("cannot duplicate schema_path");
    }
}

void ClxApiEventWriter::EnableFileWriter(bool enable) {
    api_params_.file_write_enabled = enable;
}

void ClxApiEventWriter::SetIPCSocketsDir(const char* ipc_sockets_dir) {
    if (!set_str(api_params_.ipc_sockets_dir, ipc_sockets_dir)) {
        throw std::runtime_error("cannot duplicate ipc_sockets_dir");
    }
}

void ClxApiEventWriter::EnableIPC(bool enable) {
    api_params_.ipc_enabled = true;
}

void ClxApiEventWriter::SetIPCReattachMaxInterval(uint32_t ipc_max_reattach_time_msec) {
    api_params_.ipc_max_reattach_time_msec = ipc_max_reattach_time_msec;
}

void ClxApiEventWriter::SetIPCReattachMaxTries(uint8_t ipc_max_reattach_tries) {
    api_params_.ipc_max_reattach_tries = ipc_max_reattach_tries;
}

void ClxApiEventWriter::SetIPCSocketTimeout(uint32_t ipc_socket_timeout_msec) {
    api_params_.ipc_socket_timeout_msec = ipc_socket_timeout_msec;
}

bool ClxApiEventWriter::Start(void) {
    clx_api_provider_t provider = {.version     = {{1, 0, 0, 0}},
                                   .name        = strdup("prometheus aggr"),
                                   .description = nullptr,
                                   .data        = nullptr,
                                   .initialize  = _dummy_provider_initialize,
                                   .finalize    = _dummy_provider_finalize};

    api_ctx_ = clx_api_create_context(&api_params_, &provider);
    if (!api_ctx_) {
        std::cerr << "Failed to create API context" << std::endl;
        return false;
    }

    return true;
}

void ClxApiEventWriter::Stop(void) {
    clx_api_destroy_context(api_ctx_);
    api_ctx_ = nullptr;
}

void show_usage(std::string name) {
    std::cerr << "Usage: " << name << " <option(s)"
              << "Options:\n"
              << "\t-h,--help\n\t\tShow this help message\n"
              << "\t-e,--end-point ENDPOINT_URL[,POLLING_PERIOD_MSEC]\n\t\tadd a Prometheus EndPoint to monitor\n"
              << "\t-f,--enable-file-writer\n\t\tenable CLX API file writer\n"
              << "\t--file-writer-data-root DATA_ROOT\n\t\tset CLX API file writer's data root\n"
              << "\t-i,--enable-ipc\n\t\tenable CLX API IPC\n"
              << "\t--ipc-sockets-dir IPC_SOCKET_DIR\n\t\tenable CLX API IPC with IPC_SOCKET_DIR\n"
              << "\t--ipc-sockets-timeout IPC_SOCKET_TIMEOUT\n\t\tset CLX API IPC socket timeout to IPC_SOCKET_TIMEOUT\n"
              << "\t--ipc-reattach-max-time IPC_REATTACH_MAX_TIME\n\t\tset CLX API IPC attach/reattach max interval with IPC_REATTACH_MAX_TIME\n"
              << "\t--ipc-reattach-max-tries IPC_REATTACH_MAX_TRIES\n\t\tset CLX_API IPC max reattach numer of tries with IPC_REATTACH_MAX_TRIES\n"
              << "\t-c,--max-chunk-size MAX_CHUNK_SIZE\n\t\topaque events's chunk size (0 for line-by-line split)\n"
              << "\n"
              << std::endl;
}

int parse_endpoint_val(PrometheusAggr& aggr, std::string& val) {
    std::string   url              = val;
    unsigned long polling_interval = DEFAULT_POLLING_INTERVAL;

    if (val.empty()) {
        std::cerr << "Prometheus Aggregator's endpoint address must be specified" << std::endl;
        return 1;
    }

    size_t dpos = val.find(",");
    if (dpos != std::string::npos) {
        std::string interval = val.substr(dpos + 1);

        url = val.substr(0, dpos);

        if (url.empty() || interval.empty()) {
            std::cerr << "Bad Prometheus Aggregator's endpoint: " << val << std::endl;
            return 1;
        }

        char*         endp = nullptr;
        unsigned long pi   = strtoul(interval.c_str(), &endp, 0);
        if (*endp != 0 || pi >= MAX_UINT32) {  // there was an invalid char or the value is out of range
            std::cerr << "Bad Prometheus Aggregator endpoint's polling interval: " << val << std::endl;
            return 1;
        }

        polling_interval = (uint32_t)pi;
    }

    aggr.AddEndPoint(url, polling_interval);
    std::cerr << "Prometheus Aggregator's Endpoint added: " << url << " with polling interval of " << polling_interval << " sec"
              << std::endl;

    return 0;
}

int main(int argc, char** argv) {
    int               res = 1;
    std::list<pinfo>  params;
    ClxApiEventWriter writer;
    PrometheusAggr    aggr(writer);

    for (int i = 1; i < argc; ++i) {
        char* key = argv[i];
        char* val = nullptr;
        if (i + 1 < argc && argv[i + 1][0] != '-') {
            i++;
            val = argv[i];
        }
        if (val) {
            params.push_back(std::make_pair(key, val));
        }
    }

    for (auto const& p : params) {
        std::string key = p.first;
        std::string val = p.second ? p.second : "";
        if (key == "-h" || key == "--help") {
            show_usage(argv[0]);
            return 0;
        } else if (key == "-e" || key == "--end-point") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--end-point option requires one argument" << std::endl;
                goto bad_params;
            }

            if (parse_endpoint_val(aggr, val))
                goto bad_params;
        } else if (key == "-f" || key == "--enable-file-writer") {
            writer.EnableFileWriter(true);
        } else if (key == "--file-writer-data-root") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--file-writer-data-root option requires one argument" << std::endl;
                goto bad_params;
            }

            writer.SetFileWriterMaxDataRoot(val.c_str());
        } else if (key == "-i" || key == "--enable-ipc") {
            writer.EnableIPC(true);
        } else if (key == "--ipc-sockets-dir") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--ipc-sockets-dir option requires one argument" << std::endl;
                goto bad_params;
            }

            writer.SetIPCSocketsDir(val.c_str());
        } else if (key == "--ipc-sockets-timeout") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--ipc-sockets-timeout option requires one argument" << std::endl;
                goto bad_params;
            }

            char*         endp = nullptr;
            unsigned long v    = strtoul(val.c_str(), &endp, 0);
            if (*endp != 0 || v >= MAX_UINT32) {  // there was an invalid char or the value is out of range
                std::cerr << "Bad IPC_SOCKET_TIMEOUT: " << val << std::endl;
                return 1;
            }

            writer.SetIPCSocketTimeout((uint32_t)v);
        } else if (key == "--ipc-reattach-max-time") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--ipc-reattach-max-time option requires one argument" << std::endl;
                goto bad_params;
            }
            char*         endp = nullptr;
            unsigned long v    = strtoul(val.c_str(), &endp, 0);
            if (*endp != 0 || v >= MAX_UINT32) {  // there was an invalid char or the value is out of range
                std::cerr << "Bad IPC_REATTACH_MAX_TIME: " << val << std::endl;
                return 1;
            }

            writer.SetIPCReattachMaxInterval((uint32_t)v);
        } else if (key == "--ipc-reattach-max-tries") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--ipc-reattach-max-tries option requires one argument" << std::endl;
                goto bad_params;
            }

            char*         endp = nullptr;
            unsigned long v    = strtoul(val.c_str(), &endp, 0);
            if (*endp != 0 || v >= MAX_UINT8) {  // there was an invalid char or the value is out of range
                std::cerr << "Bad IPC_REATTACH_MAX_TRIES: " << val << std::endl;
                return 1;
            }

            writer.SetIPCReattachMaxTries((uint8_t)v);
        } else if (key == "-s" || key == "--split-events") {
            if (val == "") {  // Uh-oh, there was no argument
                std::cerr << "--split-events option requires one argument" << std::endl;
                goto bad_params;
            }

            char*         endp = nullptr;
            unsigned long v    = strtoul(val.c_str(), &endp, 0);
            if (*endp != 0 || v >= MAX_UINT32) {  // there was an invalid char or the value is out of range
                std::cerr << "Bad MAX_EVENT_DATA_SIZE: " << val << std::endl;
                return 1;
            }

            aggr.SetMaxChunkSize((uint32_t)v);
        }
    }

    if (!writer.Start()) {
        std::cerr << "Cannot start writer" << std::endl;
        goto writer_start_failed;
    }

    if (!aggr.PrepareLoop()) {
        std::cerr << "Cannot prepare aggregator's loop" << std::endl;
        goto aggr_prepare_loop_failed;
    }

    aggr.DoLoop();

    aggr.CleanupLoop();

aggr_prepare_loop_failed:
    writer.Stop();
writer_start_failed:
bad_params:
    return res;
}
